;##############################################################################
;
; Copyright (C) 2010-2013, Michele Cappellari
; E-mail: cappellari_at_astro.ox.ac.uk
;
; Updated versions of the software are available from my web page
; http://purl.org/cappellari/software
;
; If you have found this software useful for your research,
; I would appreciate an acknowledgment to the use of the
; "CAP_LOESS_1D routine of Cappellari et al. (2013b), which 
; implements the univariate LOESS algorithm of Cleveland (1979)"
;
; This software is provided as is without any warranty whatsoever.
; Permission to use, for non-commercial purposes is granted.
; Permission to modify for personal or internal use is granted,
; provided this copyright and disclaimer are included unchanged
; at the beginning of the file. All other rights are reserved.
;
;##############################################################################
;+
; NAME:
;       CAP_LOESS_1D
; PURPOSE:
;       Local regression LOESS smoothing http://en.wikipedia.org/wiki/Local_regression
;       Univariate: Cleveland (1979) http://www.jstor.org/stable/2286407
;
; CALLING EXAMPLE:
;       CAP_LOESS_1D, x, y, xout, yout, DEGREE=degree, FRAC=frac, $
;           NPOINTS=npoints, /ROTATE, SIGY=sigy, WOUT=wout
;
; INPUT PARAMETERS:
;   X, Y: vectors of equal number of elements containing the
;       coordinates X and the corresponding function values Y
;       to be smoothed.
;   XOUT: vector of X coordinates corresponding to YOUT.
;       If /ROTATE is not set then XOUT=X.
;   YOUT: vector of smoothed Y values (same size as Y)
;
; KEYWORDS:
;   DEGREE: degree of the local approximation (typically 1 or 2)
;   FRAC: Fraction of points to consider in the local approximation.
;       Typical values are between ~0.2-0.8. Note that the values are
;       weighted by their distance from the point under consideration.
;       This implies that the effective fraction of points contributing
;       to a given value is much smaller that FRAC.
;   NPOINTS: Number of points to consider in the local approximation.
;       This is an alternative to giving FRAC=NPOINTS/n_elements(x).
;   ROTATE: Rotate the (X,Y) coordinates to make the X axis the axis 
;       of maximum variance. This is useful to give comparable 
;       contribution to the X and Y variables in the fit.
;   SIGY: 1-sigma errors for the Y values. If this keyword is used
;       the biweight fit is done assuming those errors. If this keyword
;       is *not* used, the biweight fit determines the errors in Y
;       from the scatter of the neigbouring points.
;   WOUT: Output weights used in the fit. This can be used to
;       identify outliers: wout=0 for outliers deviations >4sigma.
;
; HISTORY:
;   V1.0: Michele Cappellari Oxford, 15 December 2010
;   V1.1: Rescale after rotating to axis of maximum variance.
;       MC, Vicenza, 30 December 2010
;   V1.11: Fix: use ABS() for proper computation of "r". 
;       MC, Oxford, 07 March 2011
;   V1.12: Return values unchanged if FRAC=0. MC, Oxford, 25 July 2011
;   V1.2: Written 1-dim version. MC, Oxford, 9 September 2011  
;   V1.21: Included DEGREE keyword. MC, Oxford, 9 November 2011
;   V1.22: Check when outliers don't change to stop iteration.
;       MC, Oxford, 2 December 2011
;   V1.3: Optionally rotate points to have direction of maxiumum variance 
;       horizonthal before the fit. MC, Oxford, 28 March 2013
;   V1.31: Updated documentation. MC, Oxford, 16 May 2013
;   V1.32: Test whether input (X,Y) have the same size. 
;       Included NPOINTS keyword. MC, Oxford, 12 October 2013
;   V1.33: Use CAP_POLYFIT_1D, which allows for zero weights. 
;       MC, Oxford, 31 October 2013
;   V1.34: Include SIGY and WOUT keywords. Updated documentation. 
;       MC, Paranal, 7 November 2013
;-
;------------------------------------------------------------------------
pro cap_loess_1d, x1, y1, xout, yout, $
    DEBUG=debug, DEGREE=deg1, FRAC=frac1, NPOINTS=npoints1, $
    ROTATE=rotate, SIGY=sigy, WOUT=wout
compile_opt idl2
on_error, 2

frac = ~N_ELEMENTS(frac1) ? 0.3 : frac1
if frac eq 0 then begin
    yout = y
    return
endif
deg = ~N_ELEMENTS(deg1) ? 1 : deg1

n = N_ELEMENTS(x1)
if n ne n_elements(y1) then $
    message, 'Input vectors (X, Y) must have the same size'

npoints = ~n_elements(npoints1) ? ceil(n*frac) : npoints1

if KEYWORD_SET(rotate) then begin

    ; Robust calculation of the axis of maximum variance
    ;
    nsteps = 180
    angles = dindgen(nsteps)
    sig = fltarr(nsteps)
    for j=0,nsteps-1 do begin
        cap_rotate_points, x1, y1, angles[j], x2, y2
        sig[j] = robust_sigma(x2)
    endfor
    tmp = max(sig,k) ; Find index of max value
    cap_rotate_points, x1, y1, angles[k], x2, y2
    x = x2
    y = y2
    
endif else begin

    x = x1
    y = y1
    
endelse
 
yout = dblarr(n,/NOZERO)
wout = yout

for j=0,n-1 do begin

    dist = abs(x - x[j])
    w = (sort(dist))[0:npoints-1]
    distWeights = (1d - (dist[w]/dist[w[npoints-1]])^3)^3 ; tricube function distance weights
    yfit = cap_polyfit_1d(x[w], y[w], deg, WEIGHTS=distWeights)

    ; Robust fit from Sec.2 of Cleveland (1979)
    ; Use errors if those are known.
    ;
    bad = -1 ; None
    for p=0,9 do begin ; do at most 10 iterations
        if ~n_elements(sigy) then begin ; Errors are unknown
            aerr = abs(yfit - y[w]) ; Note ABS()
            mad = median(aerr) ; Characteristic scale
            uu = (aerr/(6d*mad))^2 < 1d ; For a Gaussian: sigma=1.4826*MAD
        endif else $ ; Errors are assumed known
            uu = ((yfit - y[w])/(4d*sigy[w]))^2 < 1d
        biWeights = (1d - uu)^2
        totWeights = distWeights*biWeights
        yfit = cap_polyfit_1d(x[w], y[w], deg, WEIGHTS=totWeights)
        badOld = bad
        bad = where(biWeights lt 0.34d) ; 99% confidence outliers
        if array_equal(badOld, bad) then break
    endfor

    yout[j] = yfit[0]
    wout[j] = biWeights[0]

    if keyword_set(debug) then begin
        cgplot, x, y, PSYM='plus sign'
        cgplots, x[w], y[w], psym='open diamond'
        cgplots, x[w], yfit, color='red'
        cgplots, x[w[bad]], y[w[bad]], color='green', psym='filled circle'
        wait, 0.1
    endif
    
endfor

if KEYWORD_SET(rotate) then begin

    cap_rotate_points, x, yout, -angles[k], x3, y3
    xout = x3
    yout = y3
    
endif else xout = x    

end
;------------------------------------------------------------------------

