;##############################################################################
;
; Copyright (C) 2010-2013, Michele Cappellari
; E-mail: cappellari_at_astro.ox.ac.uk
;
; Updated versions of the software are available from my web page
; http://purl.org/cappellari/software
;
; If you have found this software useful for your research,
; I would appreciate an acknowledgment to the use of the
; "CAP_LOESS_2D routine of Cappellari et al. (2013b), which implements 
; the multivariate LOESS algorithm of Cleveland & Devlin (1988)"
;
; This software is provided as is without any warranty whatsoever.
; Permission to use, for non-commercial purposes is granted.
; Permission to modify for personal or internal use is granted,
; provided this copyright and disclaimer are included unchanged
; at the beginning of the file. All other rights are reserved.
;
;##############################################################################
;+
; NAME:
;       CAP_LOESS_2D
; PURPOSE:
;       Local regression LOESS smoothing http://en.wikipedia.org/wiki/Local_regression
;       Univariate: Cleveland (1979) http://www.jstor.org/stable/2286407
;       Multivariate: Cleveland & Devlin (1988) http://www.jstor.org/stable/2289282
;
; CALLING EXAMPLE:
;       CAP_LOESS_2D, x, y, z, zout, DEGREE=degree, FRAC=frac, $
;           NPOINTS=npoints, /RESCALE, SIGZ=sigz, WOUT=wout
;           
; INPUT PARAMETERS:
;   X, Y, Z: vectors of equal number of elements containing the
;       coordinates (X,Y) and the corresponding function values Z
;       to be smoothed.
;   ZOUT: vector of smoothed Z values (same size as Z)
;
; KEYWORDS:
;   DEGREE: degree of the local approximation (typically 1 or 2)
;   FRAC: Fraction of points to consider in the local approximation.
;       Typical values are between ~0.2-0.8. Note that the values are
;       weighted by their distance from the point under consideration.
;       This implies that the effective fraction of points contributing 
;       to a given value is much smaller that FRAC.
;   NPOINTS: Number of points to consider in the local approximation.
;       This is an alternative to giving FRAC=NPOINTS/n_elements(x).
;   RESCALE: Rescale the (X,Y) coordinates to transform the inertia 
;       ellipse into a circle. Using this keyword is useful when
;       the X and Y coordinates do not represent the same physical 
;       quantity.
;   SIGZ: 1-sigma errors for the Z values. If this keyword is used
;       the biweight fit is done assuming those errors. If this keyword 
;       is *not* used, the biweight fit determines the errors in Z
;       from the scatter of the neigbouring points.  
;   WOUT: Output weights used in the fit. This can be used to 
;       identify outliers: wout=0 for outliers deviations >4sigma.
;  
; MODIFICATION HISTORY:
;   V1.0: Michele Cappellari Oxford, 15 December 2010
;   V1.1: Rescale after rotating to axis of maximum variance.
;       MC, Vicenza, 30 December 2010
;   V1.11: Fix: use ABS() for proper computation of "r". 
;       MC, Oxford, 07 March 2011
;   V1.12: Return values unchanged if FRAC=0. MC, Oxford, 25 July 2011
;   V1.13: Check when outliers don't change to stop iteration.
;       MC, Oxford, 2 December 2011
;   V1.14: Updated documentation. MC, Oxford, 16 May 2013
;   V1.32: Test whether input (X,Y,Z) have the same size. 
;       Included NPOINTS keyword. MC, Oxford, 12 October 2013
;   V1.33: Use CAP_POLYFIT_2D. Removed /QUARTIC keyword and replaced
;       by DEGREE keyword like CAP_LOESS_1D. MC, Oxford, 31 October 2013
;   V1.34: Include SIGZ and WOUT keywords. Updated documentation. 
;       MC, Paranal, 7 November 2013
;-
;------------------------------------------------------------------------
pro cap_loess_2d, x1, y1, z, zout, $
    FRAC=frac1, DEGREE=degree1, RESCALE=rescale, $
    DEBUG=debug, NPOINTS=npoints1, SIGZ=sigz, WOUT=wout
compile_opt idl2
on_error, 2    

frac = ~n_elements(frac1) ? 0.5 : frac1
if frac eq 0 then begin
    zout = z
    return
endif
degree = ~n_elements(degree1) ? 1 : degree1 ; Fit a plane by default

n = n_elements(x1)
if n ne n_elements(y1) || n ne n_elements(z) then $
    message, 'Input vectors (X, Y, Z) must have the same size' 

npoints = ~n_elements(npoints1) ? ceil(n*frac) : npoints1

if KEYWORD_SET(rescale) then begin

    ; Robust calculation of the axis of maximum variance
    ;
    nsteps = 180
    angles = dindgen(nsteps)
    sig = fltarr(nsteps)
    for j=0,nsteps-1 do begin
        cap_rotate_points, x1, y1, angles[j], x2, y2
        sig[j] = robust_sigma(x2)
    endfor
    tmp = max(sig,k) ; Find index of max value
    cap_rotate_points, x1, y1, angles[k], x2, y2
    x = (x2 - biweight_mean(x2)) / robust_sigma(x2)
    y = (y2 - biweight_mean(y2)) / robust_sigma(y2)
    
endif else begin

    x = x1
    y = y1
    
endelse  
   
zout = dblarr(n,/NOZERO)
wout = zout

for j=0,n-1 do begin

    dist = sqrt((x - x[j])^2 + (y - y[j])^2)
    w = (sort(dist))[0:npoints-1]
    distWeights = (1d - (dist[w]/dist[w[npoints-1]])^3)^3 ; tricube function distance weights
    zfit = cap_polyfit_2d(x[w], y[w], z[w], degree, WEIGHTS=distWeights)

    ; Robust fit from Sec.2 of Cleveland (1979)
    ; Use errors if those are known.
    ;
    bad = -1
    for p=0,9 do begin ; do at most 10 iterations
        if ~n_elements(sigz) then begin ; Errors are unknown
            aerr = abs(zfit - z[w]) ; Note ABS()
            mad = median(aerr) ; Characteristic scale
            uu = (aerr/(6d*mad))^2 < 1d ; For a Gaussian: sigma=1.4826*MAD 
        endif else $ ; Errors are assumed known 
            uu = ((zfit - z[w])/(4d*sigz[w]))^2 < 1d         
        biWeights = (1d - uu)^2
        totWeights = distWeights*biWeights
        zfit = cap_polyfit_2d(x[w], y[w], z[w], degree, WEIGHTS=totWeights)
        badOld = bad
        bad = where(biWeights lt 0.34d) ; 99% confidence outliers
        if array_equal(badOld, bad) then break
    endfor
    
    zout[j] = zfit[0]
    wout[j] = biWeights[0]

    if keyword_set(debug) then begin
        cgplot, x, y, PSYM='Plus sign', /YNOZERO, /ISO
        cgplot, x[w], y[w], PSYM='Open diamond', /OVERPLOT, COLOR='green'
        cgplots, x[j], y[j], PSYM='filled circle', COLOR='red'
        tvcircle, dist[w[-1]], x[j], y[j], 'red', /DATA
        wait, 0.1
    endif

endfor

end
;------------------------------------------------------------------------

